local Boat = ...

--------------------------------------------------------------------------
local HARBOUR_COMP_WATERLINE = {
	TypeName = "HARBOUR_COMP_WATERLINE",
	ParentType = "COMPONENT",
	Properties = {
--		{ Name = "BoatOffsetY", Type = "float", Default = 2.11 },
		{ Name = "FlagTerrain", Type = "integer", Default = 2 ^ OBJECT_FLAG.TERRAIN:toNumber() + 2 ^ OBJECT_FLAG.BUILDING_PART:toNumber() + 2 ^ OBJECT_FLAG.PLATFORM:toNumber()},
                { Name = "BoatSpeed", Type = "float", Default = 1.0 },
                { Name = "BoatLength", Type = "float", Default = 10.0 },
                { Name = "RandomPath", Type = "boolean", Default = false },
                { Name = "MaxRandomDistanceMove", Type = "float", Default = 50.0 },
                { Name = "MaxBoatNameFlagCompNumber", Type = "integer", Default = 0 },
	}
}

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:create()
    self._boatSequenceNumberPath = 0
    self.MaxBoatNameFlagCompNumber = 2 -- don't currently works with 0
end

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:init()
       if self.RandomPath then
            self:chooseRandomSequence()
        else
            self._initialPosition = self:getOwner():getGlobalPosition()
	    self:chooseNextFlag()
        end
        if self._boatSequenceNumberPath > 1 then
            self:getOwner():setRotationY(self._rotationY + math.pi/2)
        end
        Boat:log("FlagTerrain = " .. self.FlagTerrain .. " bit TERRAIN = " .. OBJECT_FLAG.TERRAIN:toNumber())
end

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:chooseRandomSequence()
        self._timer = math.random() * self.MaxRandomDistanceMove / self.BoatSpeed
        self._rotationY = (math.random() - 0.5) * 2 * math.pi
--        Boat:log("chooseRandomSequence: _rotationY = " .. tostring(self._rotationY))
        self._boatSequenceNumberPath = 2        
end

function HARBOUR_COMP_WATERLINE:updateBoatNextPosition(_initialSequence)
        Boat:log("updateBoatNextPosition: _initialSequence = " .. _initialSequence)
        local BoatNameFlagComp = "HARBOUR_COMP_FLAG"
        if _initialSequence < 3 then
            BoatNameFlagComp = "HARBOUR_COMP_FLAG1"
        else
            local BoatFlagNumber = _initialSequence - 1
            if BoatFlagNumber > self.MaxBoatNameFlagCompNumber then
                Boat:logWarning("No more path defined")
                return false
            end
            BoatNameFlagComp = "HARBOUR_COMP_FLAG" .. BoatFlagNumber 
        end
        Boat:log("BoatNameFlagComp = " .. BoatNameFlagComp)

        Boat:log("GlobalPosition = " .. tostring(self:getOwner():getGlobalPosition()))

        local _nextFlag = self:getLevel():find(BoatNameFlagComp, false)        
        if _nextFlag == nil then
            Boat:logWarning("No more path defined")
            return false
        end
        Boat:log("_nextFlag = " .. tostring(_nextFlag))
        if not _nextFlag.BoatIsFlagSet then
            Boat:logWarning("Flag not set")
            return false
        end
        if _nextFlag:getOwner() == nil then    
            Boat:logWarning("No flag owner defined")
            return false
        end
        Boat:log("_nextFlag:getOwner() = " .. tostring(_nextFlag:getOwner()))
        local _flagObject = _nextFlag:getOwner()
        Boat:log("_positionFlag = " .. tostring(_flagObject:getGlobalPosition()))
        self._nextPosition = _flagObject:getGlobalPosition()       
        Boat:log("_nextPosition = " .. tostring(self._nextPosition))
        return true
end


function HARBOUR_COMP_WATERLINE:updateBoatSequenceNumberPath(_initialSequence, _isNewPosition)
        Boat:log("_initialSequence = " .. _initialSequence .. " _isNewPosition = " .. tostring(_isNewPosition))
        if _isNewPosition then
            if _initialSequence < 3 then
                Boat:log("Step 3: move to next position")
                return 3
            else 
                Boat:log("Step  greater than 3: move to next position")
                return _initialSequence + 1
            end
        else
            if _initialSequence < 3 then

                Boat:log("Step 1: nothing to do")
                return 1 -- nothing to do
            else
                self._nextPosition = self._initialPosition
                Boat:log("Step 2: return to initial position")
                return 2 -- return to initial position
            end
        end
end

function HARBOUR_COMP_WATERLINE:updateBoatNextMove()
                local DeltaX = self._nextPosition[1] - self:getOwner():getGlobalPosition()[1]
                local DeltaZ = self._nextPosition[3] - self:getOwner():getGlobalPosition()[3]
                local Distance = (DeltaX*DeltaX + DeltaZ*DeltaZ) ^ 0.5
                Boat:log("angle : X = " .. DeltaX .. " Z = " .. DeltaZ .. " Distance = " .. Distance)
                if Distance > 0 then
                    if DeltaX > Distance then
                        DeltaX = 1.0
                    elseif DeltaX < -Distance then
                        DeltaX = -1.0
                    else
                        DeltaX = DeltaX / Distance
                    end 
                    self._angleX = math.asin(DeltaX)
                    if DeltaZ > Distance then
                        DeltaZ = 1.0
                    elseif DeltaZ < -Distance then
                        DeltaZ = -1.0
                    else
                        DeltaZ = DeltaZ / Distance
                    end 
                    if DeltaZ > 0 then
                        self._angleZ = self._angleX
                    else
                        self._angleZ = math.pi - self._angleX
                    end
                    self._rotationY = self._angleZ                  
                    Boat:log("self._rotationY = " .. self._rotationY .. " X = " .. self._angleX .. " Z = " .. self._angleZ)
                    self._timer = Distance / self.BoatSpeed
                end 
end

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:chooseNextFlag()
        local IsUpdatedBoatNextPosition = false
        Boat:log("IsUpdatedBoatNextPosition before = " .. tostring(IsUpdatedBoatNextPosition))
        IsUpdatedBoatNextPosition = self:updateBoatNextPosition(self._boatSequenceNumberPath)
        Boat:log("IsUpdatedBoatNextPosition after = " .. tostring(IsUpdatedBoatNextPosition))
        self._boatSequenceNumberPath = self:updateBoatSequenceNumberPath(self._boatSequenceNumberPath, IsUpdatedBoatNextPosition)
        Boat:log("Sequence after = " .. self._boatSequenceNumberPath)

        self._timer = 0.0
        self._rotationY = 0
        if self._boatSequenceNumberPath > 1 then
            self:updateBoatNextMove()
            Boat:log("self._timer = " .. self._timer)
        end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:checkWaterElevation(_boatDestination)
    local raycastResult2 = {}
--    Boat:log("initial raycastResult2 = " .. tostring(raycastResult2))
     
    -- Raycast from the building position + Y = 100
    -- to building position - y = 100
    -- only on objects with a WATER flag
    local FromPosition2 = { _boatDestination[1], _boatDestination[2]+100, _boatDestination[3] }
    local ToPosition2 = { _boatDestination[1], _boatDestination[2]-100, _boatDestination[3] }
--    Boat:log("FromPosition2 = " .. tostring(FromPosition2) .. "ToPosition2 = " .. tostring(ToPosition2))
    
--    Boat:log("final raycastResult2 = RigidBody = " .. tostring(raycastResult2["RigidBody"]) .. " Position = " .. tostring(raycastResult2["Position"]) .. " Normal = " .. tostring(raycastResult2["Normal"]))
    if not self:getLevel():rayCast(FromPosition2, ToPosition2, raycastResult2, 2 ^ OBJECT_FLAG.WATER:toNumber()) then
         Boat:log("FromPosition2 = " .. tostring(FromPosition2) .. "ToPosition2 = " .. tostring(ToPosition2))
         self._timer = -100
         Boat:logWarning("Water Mill Set: Boat NOT on water")
         return 0
    else
--        Boat:log("checkWaterElevation = " .. - _boatDestination[2] + raycastResult2["Position"][2])
         return -_boatDestination[2] + raycastResult2["Position"][2]
    end
end

--------------------------------------------------------------------------
function HARBOUR_COMP_WATERLINE:update()
    local dt = self:getLevel():getDeltaTime()
    if self._timer < 0 then

        if self.RandomPath then
            self:chooseRandomSequence()
        else
            self:chooseNextFlag()
        end          
        if self._boatSequenceNumberPath > 1 then
            self:getOwner():setRotationY(self._rotationY + math.pi/2)
        end
    end
    self._timer = self._timer - dt

    if self._boatSequenceNumberPath > 1 then
        local BoatCurrent = self:getOwner():getGlobalPosition()
--        Boat:log("BoatCurrent = " .. tostring(BoatCurrent))
        local raycastResult = {}
--        Boat:log("initial raycastResult = " .. tostring(raycastResult))
     
    -- Raycast from the last current position to the next position + BoatLength
    -- only on objects with a TERRAIN flag
        local q = self:getOwner():getGlobalOrientation()
--        Boat:log("Boat:q = " .. tostring(q) ) 
        local sinp = 2 * q[4] * q[2]
        local cosp = q[4] * q[4] - q[2] * q[2]
        local normep = cosp*cosp + sinp*sinp
        if (normep < 0.9) or (normep > 1.1)  then
            Boat:logError("getGlobalOrientation() error: " .. tostring(q))
        end
--        Boat:log("sinp =" .. sinp .. " cosp = " .. cosp)
        local BoatDestination = { self:getOwner():getGlobalPosition()[1] - cosp * self.BoatLength, 
                                  self:getOwner():getGlobalPosition()[2],
                                  self:getOwner():getGlobalPosition()[3] + sinp * self.BoatLength}
--        Boat:log("BoatDestination = " .. tostring(BoatDestination))
        if self:getLevel():rayCast(self:getOwner():getGlobalPosition(), BoatDestination, raycastResult, self.FlagTerrain) then
--            Boat:log("collision with BUILDING_PART raycastResult = RigidBody = " .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
            self._timer = -1 -- TERRAIN Reached, force new path
            return
        end
--        Boat:log("final raycastResult = RigidBody = " .. tostring(raycastResult) .. tostring(raycastResult["RigidBody"]) .. " Position = " .. tostring(raycastResult["Position"]) .. " Normal = " .. tostring(raycastResult["Normal"]))
    
        local BoatEndDestination = { self:getOwner():getGlobalPosition()[1] - cosp * dt * self.BoatSpeed, 
                                     self:getOwner():getGlobalPosition()[2],
                                     self:getOwner():getGlobalPosition()[3] + sinp * dt * self.BoatSpeed}
        self:getOwner():translate({ math.sin(self._rotationY) * dt * self.BoatSpeed, self:checkWaterElevation(BoatEndDestination), math.cos(self._rotationY) * dt * self.BoatSpeed })
        if self._timer == -100 then
            self:getOwner():destroy()
        end 
     end 

end

--------------------------------------------------------------------------
Boat:registerClass(HARBOUR_COMP_WATERLINE)
